# emacs-mode: -*- python-*-
from consts import *
from MainDisplay import MainDisplay
from MainDisplayController import MainDisplayController
from TimeDisplay import TimeDisplay
from ChannelStrip import ChannelStrip, MasterChannelStrip
from ChannelStripController import ChannelStripController
from SoftwareController import SoftwareController
from Transport import Transport
import Live
import MidiRemoteScript
class MackieControl:
    __module__ = __name__
    __doc__ = "Main class that establishes the Mackie Control <-> Live interaction. It acts \n     as a container/manager for all the Mackie Control sub-components like ChannelStrips, \n     Displays and so on.\n     Futher it is glued to Lives MidiRemoteScript C instance, which will forward some \n     notifications to us, and lets us forward some requests that are needed beside the \n     general Live API (see 'send_midi' or 'request_rebuild_midi_map').\n  "

    def __init__(self, c_instance):
        self._MackieControl__c_instance = c_instance
        self._MackieControl__components = []
        self._MackieControl__main_display = MainDisplay(self)
        self._MackieControl__components.append(self._MackieControl__main_display)
        self._MackieControl__main_display_controller = MainDisplayController(self, self._MackieControl__main_display)
        self._MackieControl__components.append(self._MackieControl__main_display_controller)
        self._MackieControl__time_display = TimeDisplay(self)
        self._MackieControl__components.append(self._MackieControl__time_display)
        self._MackieControl__software_controller = SoftwareController(self)
        self._MackieControl__components.append(self._MackieControl__software_controller)
        self._MackieControl__transport = Transport(self)
        self._MackieControl__components.append(self._MackieControl__transport)
        self._MackieControl__channel_strips = [ ChannelStrip(self, i) for i in range(NUM_CHANNEL_STRIPS) ]
        for s in self._MackieControl__channel_strips:
            self._MackieControl__components.append(s)

        self._MackieControl__master_strip = MasterChannelStrip(self)
        self._MackieControl__components.append(self._MackieControl__master_strip)
        self._MackieControl__channel_strip_controller = ChannelStripController(self, self._MackieControl__channel_strips, self._MackieControl__master_strip, self._MackieControl__main_display_controller)
        self._MackieControl__components.append(self._MackieControl__channel_strip_controller)
        self._MackieControl__shift_is_pressed = False
        self._MackieControl__option_is_pressed = False
        self._MackieControl__ctrl_is_pressed = False
        self._MackieControl__alt_is_pressed = False



    def disconnect(self):
        for c in self._MackieControl__components:
            c.destroy()




    def connect_script_instances(self, instanciated_scripts):
        """Called by the Application as soon as all scripts are initialized.
       You can connect yourself to other running scripts here, as we do it
       connect the extension modules (MackieControlXTs).
    """
        try:
            from MackieControlXT.MackieControlXT import MackieControlXT
        except:
            print 'failed to load the MackieControl XT script (might not be installed)'
        found_self = False
        right_extensions = []
        left_extensions = []
        for s in instanciated_scripts:
            if (s is self):
                found_self = True
            elif isinstance(s, MackieControlXT):
                s.set_mackie_control_main(self)
                if found_self:
                    right_extensions.append(s)
                else:
                    left_extensions.append(s)

        assert found_self
        self._MackieControl__main_display_controller.set_controller_extensions(left_extensions, right_extensions)
        self._MackieControl__channel_strip_controller.set_controller_extensions(left_extensions, right_extensions)



    def application(self):
        """returns a reference to the application that we are running in"""
        return Live.Application.get_application()



    def song(self):
        """returns a reference to the Live Song that we do interact with"""
        return self._MackieControl__c_instance.song()



    def handle(self):
        """returns a handle to the c_interface that is needed when forwarding MIDI events
       via the MIDI map
    """
        return self._MackieControl__c_instance.handle()



    def refresh_state(self):
        for c in self._MackieControl__components:
            c.refresh_state()




    def is_extension(self):
        return False



    def request_rebuild_midi_map(self):
        """ To be called from any components, as soon as their internal state changed in a 
    way, that we do need to remap the mappings that are processed directly by the 
    Live engine.
    Dont assume that the request will immediately result in a call to
    your build_midi_map function. For performance reasons this is only
    called once per GUI frame."""
        self._MackieControl__c_instance.request_rebuild_midi_map()



    def build_midi_map(self, midi_map_handle):
        """New MIDI mappings can only be set when the scripts 'build_midi_map' function 
    is invoked by our C instance sibling. Its either invoked when we have requested it 
    (see 'request_rebuild_midi_map' above) or when due to a change in Lives internal state,
    a rebuild is needed."""
        for s in self._MackieControl__channel_strips:
            s.build_midi_map(midi_map_handle)

        self._MackieControl__master_strip.build_midi_map(midi_map_handle)
        for i in range(SID_FIRST, (SID_LAST + 1)):
            if (i not in function_key_control_switch_ids):
                Live.MidiMap.forward_midi_note(self.handle(), midi_map_handle, 0, i)

        Live.MidiMap.forward_midi_cc(self.handle(), midi_map_handle, 0, JOG_WHEEL_CC_NO)



    def update_display(self):
        for c in self._MackieControl__components:
            c.on_update_display_timer()




    def send_midi(self, midi_event_bytes):
        """Use this function to send MIDI events through Live to the _real_ MIDI devices
    that this script is assigned to."""
        self._MackieControl__c_instance.send_midi(midi_event_bytes)



    def receive_midi(self, midi_bytes):
        if (((midi_bytes[0] & 240) == NOTE_ON_STATUS) or ((midi_bytes[0] & 240) == NOTE_OFF_STATUS)):
            channel = (midi_bytes[0] & 15)
            note = midi_bytes[1]
            velocity = midi_bytes[2]
            if (note in range(SID_FIRST, (SID_LAST + 1))):
                if (note in display_switch_ids):
                    self._MackieControl__handle_display_switch_ids(note, velocity)
                if (note in (channel_strip_switch_ids + fader_touch_switch_ids)):
                    for s in self._MackieControl__channel_strips:
                        s.handle_channel_strip_switch_ids(note, velocity)

                if (note in channel_strip_control_switch_ids):
                    self._MackieControl__channel_strip_controller.handle_assignment_switch_ids(note, velocity)
                if (note in function_key_control_switch_ids):
                    self._MackieControl__software_controller.handle_function_key_switch_ids(note, velocity)
                if (note in software_controls_switch_ids):
                    self._MackieControl__software_controller.handle_software_controls_switch_ids(note, velocity)
                if (note in transport_control_switch_ids):
                    self._MackieControl__transport.handle_transport_switch_ids(note, velocity)
                if (note in marker_control_switch_ids):
                    self._MackieControl__transport.handle_marker_switch_ids(note, velocity)
                if (note in jog_wheel_switch_ids):
                    self._MackieControl__transport.handle_jog_wheel_switch_ids(note, velocity)
        elif ((midi_bytes[0] & 240) == CC_STATUS):
            channel = (midi_bytes[0] & 15)
            cc_no = midi_bytes[1]
            cc_value = midi_bytes[2]
            if (cc_no == JOG_WHEEL_CC_NO):
                self._MackieControl__transport.handle_jog_wheel_rotation(cc_value)
            elif (cc_no in range(FID_PANNING_BASE, (FID_PANNING_BASE + NUM_CHANNEL_STRIPS))):
                for s in self._MackieControl__channel_strips:
                    s.handle_vpot_rotation((cc_no - FID_PANNING_BASE), cc_value)




    def can_lock_to_devices(self):
        return False



    def suggest_input_port(self):
        return ''



    def suggest_output_port(self):
        return ''



    def suggest_map_mode(self, cc_no, channel):
        result = Live.MidiMap.MapMode.absolute
        if (cc_no in range(FID_PANNING_BASE, (FID_PANNING_BASE + NUM_CHANNEL_STRIPS))):
            result = Live.MidiMap.MapMode.relative_signed_bit
        return result



    def shift_is_pressed(self):
        return self._MackieControl__shift_is_pressed



    def set_shift_is_pressed(self, pressed):
        self._MackieControl__shift_is_pressed = pressed



    def option_is_pressed(self):
        return self._MackieControl__option_is_pressed



    def set_option_is_pressed(self, pressed):
        self._MackieControl__option_is_pressed = pressed



    def control_is_pressed(self):
        return self._MackieControl__control_is_pressed



    def set_control_is_pressed(self, pressed):
        self._MackieControl__control_is_pressed = pressed



    def alt_is_pressed(self):
        return self._MackieControl__alt_is_pressed



    def set_alt_is_pressed(self, pressed):
        self._MackieControl__alt_is_pressed = pressed



    def __handle_display_switch_ids(self, switch_id, value):
        if (switch_id == SID_DISPLAY_NAME_VALUE):
            if (value == BUTTON_PRESSED):
                self._MackieControl__channel_strip_controller.toggle_meter_mode()
        elif (switch_id == SID_DISPLAY_SMPTE_BEATS):
            if (value == BUTTON_PRESSED):
                self._MackieControl__time_display.toggle_mode()




# local variables:
# tab-width: 4
